//COPY COLLISION SCRIPT

//Note: At the start of the collision hooks, a check for whether the player object has the flags
//"MF_PUSHABLE" or "MF_SCENERY", is there to discern whether or not the player object is the copied player object.

//Hash table consisting of action states
local actionstates = {
	//[S_PLAY_ROLL] = true,
	//[S_PLAY_JUMP] = true,
	[S_PLAY_SPINDASH] = true,
	[S_PLAY_GLIDE] = true,
	[S_PLAY_TWINSPIN] = true,
	[S_PLAY_MELEE] = true,
	[S_PLAY_MELEE_FINISH] = true,
	[S_PLAY_MELEE_LANDING] = true
}

//Object Collision
addHook("MobjMoveCollide", function(copy, mobj)
	if not (copy.flags & (MF_PUSHABLE|MF_SCENERY)) return end
	if mobj.flags2 & (MF2_FRET|MF2_JUSTATTACKED) return end -- Can't multi-hit foes with 2+ health
	if copy.z > mobj.z + mobj.height or copy.z < mobj.z - mobj.height return end -- Z/height checks
	local player = copy.pmo.player

	//Solid scenery collision, excludes cacti
	if (mobj.flags & (MF_SCENERY|MF_SOLID)) == (MF_SCENERY|MF_SOLID)
	and mobj.type >= 1203 and mobj.type <= 1211 return true end

	if mobj.flags & (MF_FIRE|MF_PAIN|MF_MISSILE) and mobj.type != MT_CORK -- Hazardous Objects (Don't let Fang's cork hurt the copies!)
		if player.powers[pw_super] == 0 or player.powers[pw_invulnerability] == 0
			copy.death = true
		end
	end
	if mobj.flags & (MF_ENEMY|MF_BOSS) -- Collision with foes
		if mobj.health > 0
			if actionstates[copy.state] or copy.state == S_PLAY_BOUNCE or copy.state == S_PLAY_BOUNCE_LANDING
			or player.dashmode > 3*TICRATE or player.powers[pw_super] > 0 or player.powers[pw_invulnerability] > 0
			or copy.sprite2 == SPR2_ROLL
				//P_DamageMobj(mobj, copy, copy.pmo, 1)
				if not P_IsObjectOnGround(copy)
					copy.momz = 0
					copy.momz = $+8*FRACUNIT
				end
				copy.attacked = true
			else
				copy.death = true
			end
			P_DamageMobj(mobj, copy, copy.pmo, 1)
		end
	elseif mobj.type == MT_RING or mobj.type == MT_FLINGRING -- Collision with rings
		if mobj.health > 0
			P_KillMobj(mobj, copy, copy.pmo)
			S_StartSound(mobj, mobj.info.deathsound, player)
			P_GivePlayerRings(player, 1)
		end
	elseif mobj.flags & MF_MONITOR or mobj.type == MT_EGGSHIELD -- Collision with monitors (and Egg Guard's shield)
		if mobj.health > 0
			if actionstates[copy.state] or copy.state == S_PLAY_BOUNCE or copy.state == S_PLAY_BOUNCE_LANDING
			or copy.sprite2 == SPR2_ROLL
				P_KillMobj(mobj, copy, copy.pmo)
				if not P_IsObjectOnGround(copy)
					copy.momz = 0
					copy.momz = $+8*FRACUNIT
				end
				copy.attacked = true
			else
				return true
			end
		end
	end
end, MT_PLAYER)

//Separate Collision Hook for Spikes
addHook("MobjCollide", function(mobj, copy)
	if not (copy.flags & (MF_PUSHABLE|MF_SCENERY)) or copy.death return end
	local player = copy.pmo.player
	if mobj.health > 0
		if player.dashmode > 3*TICRATE or copy.state >= 42 and copy.state <= 45 -- Dashmode and Amy's physical abuse states
			P_KillMobj(mobj, copy, copy.pmo)
		end
	end
end, MT_SPIKE)

//Collision with Walls & Solid Objects
addHook("MobjMoveBlocked", function(copy)
	if not (copy.flags & (MF_PUSHABLE|MF_SCENERY)) return end

	if P_IsObjectOnGround(copy)
		if copy.z < copy.pmo.z-64*FRACUNIT or copy.z > copy.pmo.z+64*FRACUNIT
			copy.state = S_PLAY_ROLL
			copy.momz = $+10*FRACUNIT
		end
	end
end, MT_PLAYER)

//Special Wall Collision
addHook("MobjLineCollide", function(copy, line)
	if not (copy.flags & (MF_PUSHABLE|MF_SCENERY)) return end

	if line.flags & ML_IMPASSIBLE return true end
	local secf, secb = line.frontsector, line.backsector
	if secf and secf.valid
		for rover in secf.ffloors()
			if rover.flags & FF_EXISTS
				if rover.topheight > copy.z or rover.bottomheight < copy.z
					if rover.master.special == 105 then return true end
					
					if rover.master.special == 252 or rover.master.special == 253
						BreakAndOrTrigger(rover, copy, secf)
					end
					if rover.master.special == 254 or rover.master.special == 255 or rover.master.special == 256
						if actionstates[copy.state] or copy.sprite2 == SPR2_ROLL
							if rover.master.flags & ML_NOCLIMB
								if copy.skin == "amy" or copy.skin == "knuckles"
									BreakAndOrTrigger(rover, copy, secf)
									return true
								end
							else
								BreakAndOrTrigger(rover, copy, secf)
								return true
							end
						end
					end
				end
			end
		end
	end
	if secb and secb.valid
		for rover in secb.ffloors()
			if rover.flags & FF_EXISTS
				if rover.topheight > copy.z or rover.bottomheight < copy.z
					if rover.master.special == 105 return true end
					
					if rover.master.special == 252 or rover.master.special == 253
						BreakAndOrTrigger(rover, copy, secb)
					end
					if rover.master.special == 254 or rover.master.special == 255 or rover.master.special == 256
						if actionstates[copy.state] or copy.sprite2 == SPR2_ROLL
							if rover.master.flags & ML_NOCLIMB
								if copy.skin == "amy" or copy.skin == "knuckles"
									BreakAndOrTrigger(rover, copy, secb)
									return true
								end
							else
								BreakAndOrTrigger(rover, copy, secb)
								return true
							end
						end
					end
				end
			end
		end
	end
end, MT_PLAYER)