//COPY MAIN SCRIPT

//Function for counting the number of copy objects with the MF_NOCLIP flag
local function CountNoclipCopies(mo)
	local ct = mo.copytable
	local count = 0
	for i = 1, #ct
		if not (ct[i].valid) continue end
		if not (ct[i].flags & MF_NOCLIP) continue end
		count = i
	end
	return count
end

//Function for frame number manipulation
local function FrameShifting(mobj, tics, limit)
	if mobj.frameshift == nil
		mobj.frameshift = 0
	else
		if mobj.frameshift > limit then mobj.frameshift = 0 end

		if leveltime % tics == 0
			if mobj.frameshift < limit then mobj.frameshift = $+1 else mobj.frameshift = 0 end
		end

		if mobj.frameshift <= limit then mobj.frame = mobj.frameshift end
	end
end

//For allowing certain states to be set automatically
local function AutoState(mobj)
	if mobj.pmo and mobj.pmo.valid
		local player = mobj.pmo.player
		local tics, limit
		if P_IsObjectOnGround(mobj) -- Walking
			mobj.state = S_PLAY_WALK
			if mobj.skin == "sonic" and player.powers[pw_super] > 0 -- Super Sonic Walk Change
				//Idk what sprite2 name this is so I'm just gonna reference the number I got through print testing
				mobj.sprite2 = 130
			else
				mobj.sprite2 = SPR2_WALK
			end
			if mobj.skin != "metalsonic"
				limit = 7
				tics = 4
			else
				mobj.frame = A
			end
		else -- Rolling (applies to jump roll)
			mobj.state = S_PLAY_ROLL
			mobj.sprite2 = SPR2_ROLL
			if mobj.skin == "sonic" or mobj.skin == "knuckles" or mobj.skin == "metalsonic"
				limit = 5
			elseif mobj.skin == "amy" or mobj.skin == "fang"
				limit = 3
			elseif mobj.skin == "tails"
				limit = 2
			end
			tics = 2
		end
		//print(limit, tics, mobj.frame)
		if mobj.skin != "metalsonic"	
			FrameShifting(mobj, tics, limit)
		end
	end
end

//Function for pseudo-followitem
local function FollowItemCopy(mobj)
	local tics, limit

	if mobj.skin != "metalsonic" and mobj.skin != "tails"
		if mobj.fic and mobj.fic.valid then mobj.fic.state = S_NULL end
	else
		if mobj.fic and mobj.fic.valid
			//Toggle followitem's visibility
			if mobj.flags & MF_NOCLIP
				if not (mobj.fic.flags2 & MF2_DONTDRAW) then mobj.fic.flags2 = $^^MF2_DONTDRAW end
			else
				if mobj.skin == "tails"
					if mobj.fic.flags2 & MF2_DONTDRAW then mobj.fic.flags2 = $^^MF2_DONTDRAW end
				end
			end
			//Reset followitem
			if mobj.prevskin != nil and mobj.prevskin != mobj.skin
				mobj.ficErase = true
			end
			
			if not (mobj.ficErase)
				if mobj.skin == "tails"
					mobj.fic.color = mobj.color
					mobj.fic.skin = mobj.skin
					if mobj.state != S_PLAY_RUN and mobj.state != S_PLAY_FLY
						mobj.fic.state = S_TAILSOVERLAY_STAND
						mobj.fic.sprite2 = SPR2_TAL0
						if mobj.sprite2 != SPR2_ROLL
							tics = 5
						else
							tics = 2
						end
						limit = 5
					elseif mobj.state == S_PLAY_RUN
						mobj.fic.state = S_TAILSOVERLAY_RUN
						mobj.fic.sprite2 = SPR2_TAL6
						tics = 2
						limit = 3
					elseif mobj.state == S_PLAY_FLY
						mobj.fic.state = S_TAILSOVERLAY_FLY
						mobj.fic.sprite2 = SPR2_TAL7
						tics = 1
						limit = 3
					end
					FrameShifting(mobj.fic, tics, limit)
					//print(tics, limit, mobj.fic.frameshift)
					mobj.prevskin = mobj.skin
				elseif mobj.skin == "metalsonic"
					if not (mobj.flags & MF_NOCLIP) and not (mobj.eflags & MFE_UNDERWATER)
						if mobj.state == S_PLAY_WALK or mobj.state == S_PLAY_RUN
						or mobj.state == S_PLAY_FLOAT or mobj.state == S_PLAY_FLOAT_RUN
							if mobj.fic.flags2 & MF2_DONTDRAW then mobj.fic.flags2 = $^^MF2_DONTDRAW end
							if mobj.fic.state != S_JETFUMEFLASH
								mobj.fic.scale = (leveltime % 2) and $ or FRACUNIT/2
							end
						elseif mobj.state == S_PLAY_DASH
							mobj.fic.flags2 = $^^MF2_DONTDRAW
							mobj.fic.scale = (leveltime % 2) and $ or FRACUNIT
						else
							if not (mobj.fic.flags2 & MF2_DONTDRAW) then mobj.fic.flags2 = $^^MF2_DONTDRAW end
						end
					elseif mobj.eflags & MFE_UNDERWATER
						if not (mobj.fic.flags2 & MF2_DONTDRAW) then mobj.fic.flags2 = $^^MF2_DONTDRAW end
					end
					mobj.prevskin = mobj.skin
				end
				local dist, extraheight
				mobj.fic.angle = mobj.angle
				if mobj.skin == "tails"
					dist = FRACUNIT
					extraheight = 0
				elseif mobj.skin == "metalsonic"
					dist = 16*FRACUNIT
					extraheight = 16*FRACUNIT
				end
				local xpos = mobj.x+FixedMul(dist, cos(mobj.angle-ANGLE_180))
				local ypos = mobj.y+FixedMul(dist, sin(mobj.angle-ANGLE_180))
				P_TeleportMove(mobj.fic, xpos, ypos, mobj.z+extraheight)
			else
				mobj.fic.state = S_NULL
				mobj.ficErase = false
				mobj.prevskin = nil
			end
		else
			if mobj.skin == "tails"
				mobj.ficType = MT_TAILSOVERLAY
			elseif mobj.skin == "metalsonic"
				mobj.ficType = MT_METALJETFUME
			end
			mobj.fic = P_SpawnMobj(mobj.x+mobj.momx, mobj.y+mobj.momy, mobj.z+mobj.momz, mobj.ficType)
			if mobj.skin == "metalsonic" then mobj.fic.state = S_JETFUMEFLASH end
		end
	end
end

//Metal Sonic Float & Fang Bounce Checks (for IsPlayerInAction function)
addHook("AbilitySpecial", function(player)
	if player.mo.skin == "metalsonic" then player.mo.specialfloat = true end
	if player.mo.skin == "fang" then player.mo.bouncing = true end
	return false
end)
//Super Sonic Float Check (for IsPlayerInAction function)
addHook("JumpSpinSpecial", function(player)
	if player.mo.skin == "sonic" and player.powers[pw_super] > 0 then player.mo.specialfloat = true end
	return false
end)
//Shooting With Fang and Friends :D (for IsPlayerInAction function)
addHook("SpinSpecial", function(player)
	if player.mo.skin == "fang" and not (player.mo.firedshot) and P_IsObjectOnGround(player.mo)
		player.mo.firedshot = true
	end
	return false
end)
//Hash Table (for IsPlayerInAction function)
local aerialstates = {
	[S_PLAY_GLIDE] = true,
	[S_PLAY_FLY] = true,
	[S_PLAY_FLY_TIRED] = true,
	[S_PLAY_SWIM] = true
}
local function IsPlayerInAction(player)
	local boolean
	local mo = player.mo
	if player.dashmode > 3*TICRATE or aerialstates[mo.state]
	and player.climbing == 0 or mo.specialfloat
	or mo.state >= S_PLAY_SUPER_TRANS1 and mo.state <= S_PLAY_SUPER_TRANS6
	or mo.bouncing or mo.firedshot
		boolean = true
	else
		boolean = false
	end
	//Reset specialfloat variable when not floating
	if not P_IsObjectOnGround(mo)
		if mo.skin == "sonic" and not (player.cmd.buttons & BT_USE)
		or mo.skin == "metalsonic" and not (player.cmd.buttons & BT_JUMP)
			mo.specialfloat = false
		end
	else
		mo.specialfloat = false
	end
	//Reset bouncing variable when not bouncing
	if mo.skin != "fang" or mo.bouncing and not (player.cmd.buttons & BT_JUMP)
	or mo.state != S_PLAY_BOUNCE and mo.state != S_PLAY_BOUNCE_LANDING
		mo.bouncing = false
	end
	//Reset fireshot variable when not shooting
	if not P_IsObjectOnGround(mo) or mo.skin != "fang"
	or mo.state != S_PLAY_FIRE and mo.state != S_PLAY_FIRE_FINISH
		mo.firedshot = false
	end
	return boolean
end

addHook("ThinkFrame", do
	for player in players.iterate
		if player.bot > 0 return end -- Don't make this script work for real bots
		local mo = player.mo

		//For influencing how copies are spawned before making a fresh one or replacing a deceased one
		if mo.missingcopy
			local count = CountNoclipCopies(mo)
			if count == 0 then mo.missingcopy = false end
		end

		//Button timer for determining copies' positioning
		if mo.angleset == nil
			mo.angleset = ANGLE_45
			mo.distset = 88*FRACUNIT
		end
		if mo.btimer == nil then mo.btimer = 2*TICRATE end		
		if player.cmd.buttons & (BT_CUSTOM1|BT_ATTACK|BT_FIRENORMAL) and not (mo.copyaction)
			if mo.btimer > 0 then mo.btimer = $-1 end
			if mo.btimer < 60
				mo.angleset = ANGLE_45
				mo.distset = 88*FRACUNIT
			end
		else
			if mo.btimer < 2*TICRATE
				if mo.btimer >= 60
					mo.angleset = $+ANGLE_45
					if mo.distset == 44*FRACUNIT
						mo.distset = 88*FRACUNIT
					elseif mo.distset == 88*FRACUNIT
						mo.distset = 44*FRACUNIT
					end
				end
				mo.btimer = 2*TICRATE
			end
		end

		//And now for the copy object behaviour...
		for k, copy in ipairs(mo.copytable)
			if copy.valid
				//===========================================
				//==========APPEARANCE (START)===============
				//===========================================
				if player.dashmode > 3*TICRATE
					copy.color = (leveltime % 4) and mo.color or SKINCOLOR_FLAME
				else
					copy.color = mo.color
				end
				copy.skin = mo.skin
				
				if not (copy.attacked) and not (copy.chasingplayer)
					if mo.state != S_PLAY_FLY and mo.state != S_PLAY_FLY_TIRED and mo.state != S_PLAY_SWIM
						if not (copy.sprung) and not (copy.fall)
							if mo.state == S_PLAY_SPRING or mo.state == S_PLAY_FALL and not (player.pflags & PF_JUMPED) -- Falling from spring launch not jumping
								AutoState(copy)
							else
								copy.state = mo.state
								copy.sprite2 = mo.sprite2
								copy.frame = mo.frame
							end
						end
					else
						local tics, limit
						if not (copy.eflags & MFE_UNDERWATER)
							tics = 2
							limit = 1
							copy.state = S_PLAY_FLY
						else
							tics = 4
							limit = 3
							copy.state = S_PLAY_SWIM
						end
						FrameShifting(copy, tics, limit)
					end
				elseif copy.attacked and copy.skin == "amy"
					copy.state = S_PLAY_TWINSPIN
					FrameShifting(copy, 2, 5)
				end

				if copy.spawntimer > 0
					if not (copy.flags & MF_NOCLIP) then copy.flags2 = $^^MF2_DONTDRAW end
					copy.spawntimer = $-1
				else
					if copy.flags & MF_NOCLIPTHING then copy.flags = $^^MF_NOCLIPTHING end
					if not (copy.flags & MF_NOCLIP)
						if copy.flags2 & MF2_DONTDRAW then copy.flags2 = $^^MF2_DONTDRAW end
					end
				end
				//===========================================
				//============APPEARANCE (END)===============
				//===========================================
				

				//=====================================================
				//====MOVEMENT, POSITION & STATE ADJUSTMENTS(START)====
				//=====================================================
				local xpos, ypos
				local ang = mo.angleset+(ANGLE_90*k)
				xpos = mo.x+FixedMul(mo.distset, cos(mo.angle-ang))
				ypos = mo.y+FixedMul(mo.distset, sin(mo.angle-ang))

				if player.powers[pw_carry] > 0 or player.climbing > 0
					P_TeleportMove(copy, mo.x, mo.y, mo.z)
					copy.chasingplayer = false
					copy.attacked = false
				else
					if xpos != nil and ypos != nil
						local xydist = R_PointToDist2(copy.x, copy.y, xpos, ypos)	
						local spd = player.speed
						if xydist < 1 then xydist = 1 end

						if xydist > spd+24*FRACUNIT
							copy.angle = R_PointToAngle2(copy.x, copy.y, xpos, ypos)
							copy.momx = FixedMul(FixedDiv(xpos - copy.x, xydist), 12*FRACUNIT+spd)
							copy.momy = FixedMul(FixedDiv(ypos - copy.y, xydist), 12*FRACUNIT+spd)
							copy.chasingplayer = true							
							AutoState(copy)							
						else
							copy.angle = player.drawangle
							copy.momx = xpos - copy.x + player.rmomx
							copy.momy = ypos - copy.y + player.rmomy
							copy.chasingplayer = false
						end
						//print(xydist/FRACUNIT)
					end
				end
				
				//Z-positioning when jumping
				if not P_IsObjectOnGround(mo) and player.powers[pw_carry] == 0
					if not (copy.attacked) and not (copy.chasingplayer) and player.pflags & PF_JUMPED
						copy.momz = mo.momz
					end
				elseif player.powers[pw_carry] > 0 -- when jumping from objects that carry you
					copy.sprung = false
					copy.fall = false
					copy.momz = mo.z - copy.z
					copy.angle = player.drawangle
				end
				//=====================================================
				//=====MOVEMENT, POSITION & STATE ADJUSTMENTS(END)=====
				//=====================================================
				
				
				//========================================
				//===========ABILITIES (START)============
				//========================================
				if IsPlayerInAction(player) and not (copy.flags & MF_NOCLIP) 
					if copy.skin != "fang"
						//Tails' tails chopper sound
						if copy.skin == "tails" and not (copy.eflags & MFE_UNDERWATER)
							if copy.choppertimer == nil
								copy.choppertimer = 0
							elseif copy.choppertimer < 4
								copy.choppertimer = $+1
							else
								copy.choppertimer = 0
								S_StartSound(copy, sfx_putput)
							end
						end
						
						//Height adjustments when using aerial abilities (e.g. flight & glide)
						if copy.z > mo.z + 16*FRACUNIT
							copy.momz = $-FRACUNIT
						elseif copy.z < mo.z - 16*FRACUNIT
							copy.momz = $+FRACUNIT
						else
							copy.momz = mo.z - copy.z + mo.momz
						end

					//Are you "Fang"? Then do your own thing.
					else
						//Bounce Bounce Stuff
						if mo.bouncing
							if copy.floortimer == nil then copy.floortimer = 0 end
							if not P_IsObjectOnGround(copy)
								copy.state = S_PLAY_BOUNCE
								copy.floortimer = 0
							else
								copy.state = S_PLAY_BOUNCE_LANDING
								if copy.floortimer < 3
									copy.floortimer = $+1
								else
									S_StartSound(copy, sfx_boingf)
									copy.momz = $+16*FRACUNIT
								end
							end
						end
						//Shooty Shooty Stuff
						if mo.firedshot and not (mo.bouncing) and copy.spawntimer <= 0
							copy.angle = R_PointToAngle2(mo.x, mo.y, copy.x, copy.y)
							if not copy.shooting
								copy.shooting = true
								local x, y = FixedMul(16*FRACUNIT, cos(copy.angle)), FixedMul(16*FRACUNIT, sin(copy.angle))
								local bullet = P_SpawnMobj(copy.x+x, copy.y+y, copy.z+16*FRACUNIT, MT_CORK)
								P_InstaThrust(bullet, copy.angle, 48*FRACUNIT)
							end
						end
					end
				end
				//Make Fang copies shoot again (fix - leave outside of "IsPlayerInAction" block)
				if not (mo.firedshot) then copy.shooting = false end
				//======================================
				//===========ABILITIES (END)============
				//======================================
				
				
				//============================================
				//===========MISCELLANEOUS (START)============
				//============================================

				//Follow Item
				FollowItemCopy(copy)
				
				//Teleport clone if they're too far away
				if copy.teleporttimer == nil
					copy.teleporttimer = 5*TICRATE/2
				else
					if not (copy.flags & MF_NOCLIP)
						//Distance checking - teleport clone to player under certain conditions
						local cdist = R_PointToDist2(copy.x, copy.y, mo.x, mo.y)
						if cdist > 256*FRACUNIT or copy.z > mo.z+128*FRACUNIT or copy.z < mo.z-128*FRACUNIT
							if copy.teleporttimer > 0
								copy.teleporttimer = $-1
							else
								P_TeleportMove(copy, xpos, ypos, mo.z+mo.momz)
								copy.spawntimer = 18
								copy.flags = $|MF_NOCLIPTHING
								copy.teleporttimer = 5*TICRATE/2
							end
						else
							copy.teleporttimer = 5*TICRATE/2
						end
					end
				end
				
				copy.height = mo.height -- For spin height adjustments
				copy.pmo = mo -- store player.mo for future reference
				
				if player.exiting != 0 -- Player exiting check
					if not (copy.flags & MF_NOCLIP)
						P_AddPlayerScore(player, 1000*k)
						local score = P_SpawnMobj(copy.x, copy.y, copy.z+32*FRACUNIT, MT_SCORE)
						score.frame = D
					end
					copy.death = true
				end
				if player.playerstate & PST_DEAD then copy.death = true end -- If player dies, then copies should die too.

				if copy.attacked and P_IsObjectOnGround(copy) -- Resetting attack variable
					copy.attacked = false
				end
				
				//Spring Interaction
				if not (mo.bouncing)
					if copy.eflags & MFE_SPRUNG then copy.sprung = true end
					if copy.sprung
						if not P_IsObjectOnGround(copy)
							copy.state = S_PLAY_SPRING
							copy.attacked = false
							if copy.momz <= 0
								copy.sprung = false
								copy.fall = true
							end
						else
							copy.sprung = false
						end
					elseif copy.fall
						local tics, limit = 3, 1
						copy.state = S_PLAY_FALL
						FrameShifting(copy, tics, limit)
					end
					if copy.fall and P_IsObjectOnGround(copy)
						copy.fall = false
					end
				end
				
				//Breaking FOFs from above or below with Fang or Amy
				if P_IsObjectOnGround(copy) and not (copy.flags & MF_NOCLIP)
					if copy.state == S_PLAY_BOUNCE or copy.state == S_PLAY_BOUNCE_LANDING
					or copy.state == S_PLAY_MELEE or copy.state == S_PLAY_MELEE_FINISH
					or copy.state == S_PLAY_TWINSPIN
						local sector = copy.subsector.sector
						for rover in sector.ffloors()
							if not (rover.flags & FF_EXISTS) continue end
							if rover.master.special < 254 or rover.master.special > 256 continue end
							if not (copy.eflags & MFE_VERTICALFLIP)
								if copy.z != rover.topheight continue end
								BreakAndOrTrigger(rover, copy, sector)
							else
								if copy.z != rover.bottomheight continue end
								BreakAndOrTrigger(rover, copy, sector)
							end
						end
					end
				end
				
				//Spawning invincibility sparkles and afterimages of the clones
				if not (copy.flags & MF_NOCLIP) 
					if player.powers[pw_invulnerability] > 0
						if leveltime % 8 == 0
							P_SpawnMobj(copy.x+copy.momx, copy.y+copy.momy, copy.z+copy.momz, MT_IVSP)
						end
					end	
					if player.powers[pw_sneakers] > 0
						local ghost = P_SpawnGhostMobj(copy)
						ghost.fuse = 2
					end
				end		
				//==========================================
				//===========MISCELLANEOUS (END)============
				//==========================================

				//===============DEATH (actually fake but who cares)=================
				if copy.death and not (copy.flags & MF_NOCLIP) -- Death
					P_SpawnMobj(copy.x, copy.y, copy.z, MT_EXPLODE)
					S_StartSound(copy, sfx_pop)
					copy.flags2 = $|MF2_DONTDRAW
					copy.flags = ($|(MF_NOCLIP|MF_SCENERY)) & ~(MF_PUSHABLE)
					copy.key = k
					mo.missingcopy = true
					copy.death = false
				end
			else
				if not (player.playerstate & PST_DEAD) and player.exiting == 0
					//Copies can still be killed/removed by death pits, so let's spawn a new one (in its fake death state)
					local deadkey = k
					table.remove(mo.copytable, k)
					
					local c = P_SpawnMobj(mo.x+mo.momx, mo.y+mo.momy, mo.z+mo.momz, MT_PLAYER)
					c.color = mo.color
					c.state = mo.state
					c.skin = mo.skin
					c.sprite2 = mo.sprite2
					c.flags2 = $|MF2_DONTDRAW
					c.flags = ($|(MF_NOCLIP|MF_SCENERY)) & ~(MF_SOLID|MF_SHOOTABLE)
					c.key = deadkey
					mo.missingcopy = true
					c.spawntimer = 0
					table.insert(mo.copytable, deadkey, c)
				end
			end
		end
	end
end)